/*
 * calcrestore - a command line utility to restore a set of variables to
 * a TI89 Titanium calculator over the USB link
 *
 * Copyright (C) 2010 Michael Jordan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "ti_usb.h"
#include "ti_file.h"
#include <unistd.h>

int main(int argc, char *argv[]) {
	int return_code = EX_OK;
	bool verbose = false;
	if(argc > 1) {
		string argument;
		argument.append(argv[1]);
		bool help = false;
		bool valid = false;
		if((argument == "-h") || (argument == "--help")) {
			valid = true;
			help = true;
		}
		if((argument == "-v") || (argument == "--verbose")) {
			valid = true;
			verbose = true;
		}
		if((argument == "-q") || (argument == "--quiet")) {
			valid = true;
		}
		if(!valid) help = true;
		if(help) {
			cout <<
					"\ncalcrestore is a utility to restore files to a\n" <<
					"TI89 Titanium calculator over the USB link.\n\n" << flush;
			cout <<
					"It will get a listing of the valid variable files in the\n" <<
					"current directory, and then begin putting the variables onto\n" <<
					"the calculator.  The usual way of using this utility will\n" <<
					"be to dump the assorted variable files into a directory,\n" <<
					"change to that directory and then run this utility.\n" <<
					"This utility runs in two basic modes - user interface and script.\n" <<
					"When using script mode the utility will go about its business\n" <<
					"quietly and produce NO output to stdout.\n" <<
					"This is provided for use in scripts which will check the return\n"
					"value.\n" <<
					"The user interface mode outputs information at each step of the\n" <<
					"process so that a user can see the process progress and note\n"
					"any warnings as they occur.\n\n" <<
					"Command line arguments:\n" <<
					"\t-h, --help - These options will print this help screen\n" <<
					"\t\tthen exit.\n" <<
					"\t-v, --verbose - These options will put the program into\n" <<
					"\t\tuser interface mode.\n" <<
					"\t-q, --quiet - These options will put the program into\n" <<
					"\t\tscript mode.\n" <<
					"\nIn either case this utility will return one of three values:\n" <<
					"\tEX_OK\t\t0\tThe process completed\n" <<
					"\t\t\t\tsuccessfully and all vars were\n" <<
					"\t\t\t\tloaded.\n" <<
					"\tENOTCONN\t107\tNo calculator was located on the\n" <<
					"\t\t\t\tUSB link.\n" <<
					"\tEIO\t\t5\tAt least one of the vars was not\n" <<
					"\t\t\t\tloaded successfully to the calculator.\n" <<
					flush;
			cout <<
					"Keep in mind that this is, in fact, free software, developed\n" <<
					"independently and therefore comes with no warranty.\n" << flush;
			return return_code;
		}
	} else {
		return_code = EX_USAGE;
		cout << "This utility requires one of the following arguments:\n" <<
				"\t-h, --help\n" <<
				"Or one of the following:\n" <<
				"\t-v\n" <<
				"\t-q\n" <<
				"Alternatively -v may be --verbose or -q may be --quiet\n" << flush;
		return return_code;
	}

	if(verbose) cout << "Searching for calculator...\n" << flush;
	ti_usb calc;

	// Calling is_good() here verifies that there was a calculator located on
	// the USB bus and that we can in fact communicate with it.
	if(calc.is_good()) {
		vector<ti_usb::ls_item_info> items;
		ti_usb::ls_item_info ls_item;

		if(verbose) cout << "Calculator located...\nGetting listing...\n";
		calc.ls(items);

		string dir = ".";
		vector<string> files;

		// Get a listing of all the files in the current directory - start by
		// eliminating any '.' (dot) files in the listing as well as any
		// directories.
		if(verbose) cout << "Getting system directory listing and eliminating dot files\n";
		calc.get_sys_directory(dir, files);
		struct stat st;
		int ret;
		for(int i = 0; i < (int)files.size(); i++) {
			ret = stat(files[i].c_str(), &st);
			if((files[i].c_str()[0] == '.') ||
					(S_ISDIR(st.st_mode))) files.erase(files.begin() + i);
		}

		// Now attempt to open each file by creating a ti_file object for each.
		// Any that are not understandable to us we eliminate from the list.
		if(verbose) cout << "Testing each file and eliminating each invalid\n";
		string filename;
		string cwd;
		cwd.append(getcwd(NULL, 0));
		cwd += "/";
		for(int i = 0; i < (int)files.size(); i++) {
			filename = cwd + files[i];
			ti_file *file = new ti_file(filename);
			if(!file->is_good()) {
				files.erase(files.begin() + i);
				i--;
			}
			delete file;
		}

		// Get the amount of free storage and begin iterating through our
		// remaining entries to determine:
		// 1.  Is the var already on the calculator - determined by a match
		//     in the folder as well as the name of the var
		// 2.  Is there enough storage available - determined by comparing
		//     var size to free ram if it is going into ram or by comparing
		//     var size to free flash if it is going into flash memory
		if(verbose) cout << "Installing valid vars\n";
		int free_ram = calc.get_free_ram();
		int free_flash = calc.get_free_flash();
		bool is_on_calc;
		bool enough_room;

		// Iterate through our list of files.
		for(int i = 0; i < (int)files.size(); i++) {
			filename = cwd + files[i];
			ti_file curr_file(filename);

			// Iterate through each entry within a file.
			for(int j = 0; j < curr_file.num_entries(); j++) {
				string curr_name = curr_file.name(j);
				string curr_folder = curr_file.folder(j);
				is_on_calc = false;

				// Iterate through the list of items already installed on the
				// calculator.
				for(int k = 0; k < (int)items.size(); k++) {
					is_on_calc = ((curr_name == items[k].name) &&
							(curr_folder == items[k].parent_folder));
					if(is_on_calc) {

						// Folder name and var name both match - the var is
						// already installed on the calculator (or possibly
						// one with the same name in the same folder - either
						// way we cannot install this one) - abort and go on to
						// the next var.
						if(verbose) cout << "\'" << curr_name << "\' is already on the calculator - skipping\n";
						j++;
						k = (int)items.size();
					}
				}

				// Check to see if we have enough room on the calculator for
				// the var.
				enough_room = true;
				if(curr_file.is_flash()) enough_room = (curr_file.data_size(j) < free_flash);
				else enough_room = (curr_file.data_size(j) < free_ram);
				if(enough_room && !is_on_calc) {

					// It appears that we have enough room and the var is not
					// already on the calculator.  Now let's try and send the
					// var to the calculator.
					unsigned char buf[curr_file.data_size(j)];
					curr_file.data(j, buf);
					if(verbose) cout << "Putting var \'" << curr_file.name(j) << "\' to the calculator ";
					ti_usb::ls_item_info curr_item;
					curr_item.parent_folder = curr_file.folder(j);
					curr_item.name = curr_file.name(j);
					curr_item.type = curr_file.type_id(j);
					curr_item.size = curr_file.data_size(j);
					curr_item.archived = curr_file.archived(j);
					curr_item.locked = curr_file.locked(j);
					if(calc.put_var(curr_item, buf) != 0) {
						return_code = EIO;
						if(verbose) cout << "failed\n";
					}
					else {

						// If put_var() was successful (indicated by a 0 return
						// value) we now subtract the size of the var from our
						// cached free ram/flash storage values to keep track
						// of how much storage we have left on the calculator.
						if(curr_file.is_flash()) free_flash -= curr_file.data_size(j);
						else free_ram -= curr_file.data_size(j);
						if(verbose) cout << "succeeded\n";
					}
				}
			}
		}
		return return_code;
	}
	if(verbose) cout << "No calculator found on the USB bus\n";
	return_code = ENOTCONN;
	return return_code;
}
